/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZABSTRACTTHREADLOCAL_H__
#define __ZABSTRACTTHREADLOCAL_H__

#include "zthread/ZThread.h"
#include "zthread/SynchronizationExceptions.h"

namespace ZThread {


/**
 * @class AbstractThreadLocal
 *
 * @author Eric Crahen
 * @version 1.3.5
 * @date 04-22-2001
 *
 * This class implements a common set of functionality for thread specific 
 * value classes. By default, on most systems, the native libraries only
 * provide support for void* value and no support for inhertiable or initial
 * value setting.
 *
 * This class implements the minimum set of functionality needed to provide
 * more specific, and strongly typed support than is otherwise available.
 * 
 * NOTE: 
 * Any AbstractThreadLocal MUST be a static variable. This allows for the 
 * fastest and most logical use of thread sepcific variables. In addition,
 * it is not worth the extra overhead involved with keeping track of variables
 * that may go out of scope of be deteled from one thread while others are 
 * using. While this would be possible, there is a speed tradeoff & a strange 
 * relationship between a thread and ThreadLocal variables that might 
 * disappear. It seems a little too awkawrd with very little, if any,
 * benefit to use these variables in that way. 
 *
 * @see ThreadLocal
 * @see InheritableThreadLocal
 * @see FastThreadLocal
 * 
 */
class ZTHREAD_API AbstractThreadLocal {

  friend class ThreadLocalImpl;
  friend class ThreadImpl;

  ThreadLocalImpl* _impl;

  // Disabled 
  AbstractThreadLocal(const AbstractThreadLocal&);
  AbstractThreadLocal& operator=(const AbstractThreadLocal&);

  public:

  /**
   * Create a newAbstractThreadLocal object
   */
  AbstractThreadLocal() throw();

  /**
   * Destroy this AbstractThreadLocal object
   */
  virtual ~AbstractThreadLocal() throw();

  protected:

  /** 
   * Invoked by the framework whenever a per/thread, or thread-specfic initialization
   * is required.
   *
   * @param void* - related initialization value
   */
  virtual void* initialValue(void*) throw() = 0;

  /** 
   * Invoked by the framework whenever a per/thread, or thread-specfic initialization
   * is required.
   *
   * @param void* - related initialization value
   */
  inline virtual void* childValue(void* parentValue) throw() {
    return (void*)0;
  }


  /** 
   * Invoked by the framework whenever it needs to be determined wether or not 
   * this value should propogate to child threads.
   *
   * @param bool - true if this value should propgate
   */
  virtual bool propogateValue() throw() = 0;

  /**
   * Get a value with thisAbstractThreadLocal object. Any value retrieved from
   * this object will have been set from the same Thread. No initilization is
   * performed with this function.
   *
   * @exception Synchronization_Exception - thrown if there is an error
   * allocating native thread local storage
   */
  void* fetch() const
    throw(Synchronization_Exception);

  /**
   * Get a value with thisAbstractThreadLocal object. Any value retrieved from
   * this object will have been set from the same Thread. Causes initialize() 
   * to be called if appropriate.
   *
   * @exception Synchronization_Exception - thrown if there is an error
   * allocating native thread local storage
   */
  void* get() const
    throw(Synchronization_Exception);
  
  /**
   * Set a value with this AbstractThreadLocal object. This value can only
   * be retrieved from this AbstractThreadLocal object from the same Thread
   * that set it.
   *
   * @exception Synchronization_Exception - thrown if there is an error
   * allocating native thread local storage
   */
  void* set(void*) const
    throw(Synchronization_Exception);

};


};

#endif

