/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZCHECKEDMUTEX_H__
#define __ZCHECKEDMUTEX_H__

#include "zthread/ZThread.h"
#include "zthread/Lockable.h"

namespace ZThread { 
  

class CheckedMutexImpl;

/**
 * @class CheckedMutex
 *
 * @author Eric Crahen
 * @version 1.3.4
 * @date 04-12-2001
 *
 * This is an error checking Mutex implementation. Any blocking caused
 * by this mutex can be interrupted. This mutex can also not be deadlocked
 * The performance of this mutex is still fast, but if you really need
 * speed you may want to consider using a FastMutex.
 *
 * Each method is an interruption point
 * Each method will throw a meaningful exception
 *
 * You should carefully handle exceptions thrown by this mutex. 
 */
class ZTHREAD_API CheckedMutex : public Lockable {
  
  CheckedMutexImpl* _impl;
  
  // Disable copy constructors
  CheckedMutex(const CheckedMutex&);
  CheckedMutex& operator=(const CheckedMutex&);
  
  public:

  /**
   * Create a new CheckedMutex.
   *
   * @exception Initialization_Exception - thrown if resources
   * could not be allocated
   */
  CheckedMutex() 
    throw(Synchronization_Exception);

  /**
   * Destroy this Mutex
   */
  virtual ~CheckedMutex()
    throw();

  /**
   * Lock the mutex and prevent re-entrance on the protected block of code.
   * Calling thread is blocked until the Mutex is released by the thread 
   * currently holding its lock
   *
   * @see Loackable::acquire()
   *
   * @exception Interrupted_Exception - thrown if the calling thread was
   * interrupted prior to this call, or during the time it spends waiting
   * to acquire the mutex
   * @exception Deadlock_Exception - thrown if an attempt to acquire this
   * mutex twice with the same thread is made
   * @exception Unexpected_Exception - the calling thread was killed and
   * can't be blocked
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   */
  virtual void acquire() 
    throw(Synchronization_Exception);
  
  /**
   * Lock the mutex and prevent re-entrance on the protected block of code.
   * If the given amount of time expires or the mutex is locked this function
   * returns
   *
   * see Loackable::tryAcquire()
   *
   * @exception Interrupted_Exception - thrown if the calling thread was
   * interrupted prior to this call, or during the time it spends waiting
   * to acquire the mutex
   * @exception Deadlock_Exception - thrown if an attempt to acquire this
   * mutex twice with the same thread is made
   * @exception Unexpected_Exception - the calling thread was killed and
   * can't be blocked
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   *
   * @return bool indicating the success of the lock attempt
   */
  virtual bool tryAcquire(unsigned long) 
    throw(Synchronization_Exception);
  
  /**
   * Unlock the mutex and prevent re-entrance on the protected block of code
   *
   * see Loackable::release()
   *
   * @post Exceptions are thrown after the release() is performed internally
   *
   * @exception Interrupted_Exception - thrown if the calling thread was
   * interrupted prior to this call, or during the time it spends waiting
   * to acquire the mutex
   * @exception InvalidOp_Exception - thrown if an attempt to release this
   * mutex from a thread that does not own the lock is made
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   *
   */
  virtual void release() 
    throw(Synchronization_Exception);
  
  };
  
};

#endif
