/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZCONDITION_H__
#define __ZCONDITION_H__

#include "zthread/ZThread.h"
#include "zthread/Lockable.h"

namespace ZThread {
  
// Forward declarations
class ConditionImpl;

/**
 * @class Condition
 *
 * @author Eric Crahen
 * @version 1.3.4
 * @date 04-12-2001
 *
 * Condition variables can be used to block threads until a condition has 
 * been signaled. This is very much like a POSIX condition variable.
 * Threads wait() ing on this object are blocked until they receive signal.
 *
 * A Condition variable must be used in conjunction with a Lockable object
 * The lockable object should be locked prior to calling one of the wait()
 * functions. That object will be released while the caller is blocked. When
 * the thread that has been blocked by the condition receives a signal that
 * lockable object will be locked once again before the wait() returns.
 *
 */
class ZTHREAD_API Condition {
  
  // Disable copy constructors
  Condition(const Condition&);
  Condition& operator=(const Condition&);
  
  ConditionImpl* _impl;
  
  public:
  
  /**
   * Create a new condition variable
   *
   * @exception Initialization_Exception thrown if resources to 
   * allocate the object could not be allocated
   */
  Condition() 
    throw(Synchronization_Exception);

  /**
   * Destroy Condition object
   */
  ~Condition()
    throw(); 
  
  /**
   * Wake ONE thread waiting on this condition variable.
   *
   * @exception Synchronization_Exception thrown on a general error with the
   * underlying implementation object   
   */
  void signal() 
    throw(Synchronization_Exception);

  /**
   * Wake ALL threads waiting on this condition variable.
   *
   * @exception Synchronization_Exception thrown on a general error with the
   * underlying implementation object   
   */
  void broadcast() 
    throw(Synchronization_Exception);


  /**
   * Wait for this Condition object to be signaled.
   *
   * @return bool indication the success of this operation
   *
   * @param Lockable& reference to a locked Lockable object
   *
   * @exception Interrupted_Exception - thrown if a thread blocking
   * on this object is interrupted
   *
   * @exception Synchronization_Exception thrown on a general error with the
   * underlying implementation object
   */
  void wait(Lockable&) 
    throw(Synchronization_Exception);   

  /**
   * Wait for this Condition object to be signaled for at most 
   * 'ms' milliseconds. 
   *
   * @return bool indicating the success of this operation
   *
   * @param Lockable& reference to a locked Lockable object
   * @param unsigned long milliseconds to wait on for a signal()
   *
   * @exception Interrupted_Exception - thrown if a thread blocking
   * on this object is interrupted
   * @exception Unexpected_Exception - thrown if the calling thread was 
   * killed can can't block
   * @exception Synchronization_Exception thrown on a general error with the
   * underlying implementation object
   */
  bool wait(Lockable&, unsigned long) 
    throw(Synchronization_Exception);
  
}; /* Condition */
  
};

#endif
