/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZCOUNTINGSEMAPHORE_H__
#define __ZCOUNTINGSEMAPHORE_H__

#include "zthread/ZThread.h"
#include "zthread/Lockable.h"

namespace ZThread {
  
class CountingSemaphoreImpl;
  
/**
 * @class CountingSemaphore
 *
 * @author Eric Crahen
 * @date 04-12-2001
 * @version 1.3.4
 *
 * Very similar to the Semaphore object. The difference here is that
 * there is no upper bound on this Semaphore and it will not throw
 * exceptions because a range has been exceeded
 *
 * @see Semaphore
 */
class ZTHREAD_API CountingSemaphore : public Lockable {
  
  CountingSemaphoreImpl* _impl;
  
  // Disable copy constructors
  CountingSemaphore(const CountingSemaphore&);
  CountingSemaphore& operator = (const CountingSemaphore&);
  
  public:

  /**
   * Create a new semaphore of a given size with a given count
   *
   * @param initialCount initial count to assign this semaphore
   */ 
  CountingSemaphore(int initialCount = 0) 
    throw(Synchronization_Exception);

  /**
   * Destroy this Semaphore
   */
  virtual ~CountingSemaphore() throw();

  /**
   * @see acquire() - provided to reflect the traditional Semaphore
   * semantics
   */ 
  void wait() 
    throw(Synchronization_Exception);

  /**
   * @see tryAcquire() - provided to reflect the traditional Semaphore
   * semantics
   */
  bool tryWait(unsigned long) 
    throw(Synchronization_Exception);


  /**
   * This method will increment the semaphore count by one, if any
   * threads were wait()ing on this semaphore - then one will be 
   * unblocked.
   *
   * An exception will not be thrown because a range has been exceeded
   * 
   */
  void post() 
    throw(Synchronization_Exception);
  
  /**
   * Current count of this Semaphore
   *
   * @return int - current count
   */
  virtual int count() 
    throw();    
  
  /**
   * Decrements the semaphore. If the count is 0 - block until its raised
   * above 0.
   *
   * @param unsigned long - amount of time to spend trying to acquire
   *
   * @return true - the semaphore was acquired in time, false if not
   *
   * @exception Interrupted_Exception - thrown if the thread is interrupted
   * while waiting for this semaphore, or if it was interrupted before 
   * the function was called
   * @exception Unexpected_Exception - the calling thread was killed and
   * can't be blocked
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   *
   */
  virtual bool tryAcquire(unsigned long) 
    throw(Synchronization_Exception);

  /**
   * Decrements the semaphore. If the count is 0 - block until its raised
   * above 0.
   *
   * @exception Interrupted_Exception - thrown if the thread is interrupted
   * while waiting for this semaphore, or if it was interrupted before 
   * the function was called
   * @exception Unexpected_Exception - the calling thread was killed and
   * can't be blocked
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   *
   */
  virtual void acquire() 
    throw(Synchronization_Exception);

  /**
   * Increment the semaphore
   *
   * @exception InvalidOp_Exception - thrown if this would increase the
   * count beyond the maximum
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   */
  virtual void release() 
    throw(Synchronization_Exception);

  
}; /* CountingSemaphore */


};

#endif
