/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZFASTTHREADLOCAL_H__
#define __ZFASTTHREADLOCAL_H__

#include "zthread/ZThread.h"
#include "zthread/AbstractThreadLocal.h"

namespace ZThread {

/**
 * @class FastThreadLocal
 *
 * @author Eric Crahen
 * @version 1.3.5
 * @date 04-22-2001
 *
 * This is very similar to the ThreadLocal object in function and purpose.
 * The major difference is that a FastThreadLocal never perfoms any kind
 * of context check to determine when it was first used. Because of this 
 * values are never proprogated to children threads, and the initValue()
 * method is never called by the framework. 
 */
template <class T>
class ZTHREAD_API FastThreadLocal : private AbstractThreadLocal {
  public:

  /**
   * Create a new FastThreadLocal object
   */
  inline FastThreadLocal() throw() {
  }

  /**
   * Destroy this FastThreadLocal object
   */
  inline virtual ~FastThreadLocal() throw() {
  }

  /**
   * Get a value with this ThreadLocal object. Any value retrieved from
   * this object will have been set from the same Thread. 
   *
   * @exception Synchronization_Exception - thrown if there is an error
   * allocating native thread local storage
   */
  inline T get() const
    throw(Synchronization_Exception) {

    return reinterpret_cast<T>( AbstractThreadLocal::fetch() );

  }
  
  /**
   * Set a value with this ThreadLocal object. This value can only
   * be retrieved from this ThreadLocal object from the same Thread
   * that set it.
   *
   * @exception Synchronization_Exception - thrown if there is an error
   * allocating native thread local storage
   */
  inline T set(T val) const
    throw(Synchronization_Exception) {

    return reinterpret_cast<T>( AbstractThreadLocal::set((void*)val) );

  }

  protected:

  /**
   * Never invoked by the framework.
   */
  inline virtual void* initialValue(void* initValue) throw() {   
    return initValue;
  }

  /**
   * Never invoked by the framework.
   */
  inline virtual bool propogateValue() throw() {
    return false;
  }


};


};

#endif
