/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZINHERITABLETHREADLOCAL_H__
#define __ZINHERITABLETHREADLOCAL_H__

#include "zthread/ZThread.h"
#include "zthread/ThreadLocal.h"

namespace ZThread {

/**
 * @class InheritableThreadLocal
 *
 * @author Eric Crahen
 * @version 1.4.1
 * @date 04-28-2001
 *
 * This object acts exactly like its super class ThreadLocal with one subtle
 * difference. That difference is that parent threads values can atomatically
 * be propogated to child threads when they are created (eg run() has been 
 * invoked). This is done through the initialize method. 
 */
template <class T>
class ZTHREAD_API InheritableThreadLocal : public ThreadLocal<T> {
  public:

  /**
   * Create a new InheritableThreadLocal object
   */
  inline InheritableThreadLocal() throw() {
  }

  /**
   * Destroy this InheritableThreadLocal object
   */
  inline virtual ~InheritableThreadLocal() throw() {
  }


  protected:

  /**
   * This method will be invoked by the framework when a child thread is 
   * created, if its parent has set a value for this key. This is invoked  
   * from the childs context, and will be passed the parents value.
   *
   * @param void* parentValue - parents value for this variable
   * @return void* value to set for this thread
   *
   * @pre The parent must have set a value for this to be invoked. Otherwise
   * it will not be called and the client can expect the initialValue() method
   * to be invoked as it would in the ThreadLocal class
   *
   * @post if this function has been called by the framework, initialValue()
   * will not be called
   */
  inline virtual void* childValue(void* parentValue) throw() {   
    return parentValue;
  }

  /**
   * Inform the framework this value does propogate to child threads
   *
   * @return true - always
   */
  inline virtual bool propogateValue() throw() {
    return true;
  }

};


};

#endif
