/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZMUTEX_H__
#define __ZMUTEX_H__

#include "zthread/ZThread.h"
#include "zthread/Lockable.h"

namespace ZThread { 
  
class MutexImpl;

/**
 * @class Mutex
 *
 * @author Eric Crahen
 * @version 1.5.0
 * @date 07-31-2001
 *
 * Basic MUTual EXclusion object. This mutex will perform checks before
 * a wait is entered to see if the thread has been interrupted. However,
 * once a thread is blocked by this mutex it can't be interrupted. 
 *
 * This classes behavoir is very similar to the CheckedMutex except that it
 * will not check for deadlock.
 */
class ZTHREAD_API Mutex : public Lockable {
  
  MutexImpl* _impl;
  
  // Disable copy constructors
  Mutex(const Mutex&);
  Mutex& operator = (const Mutex&);
  
  public:

  /**
   * Allocate a new mutex
   *
   * @exception Initialization_Exception - thrown if there are not
   * enough available resources
   */
  Mutex() 
    throw(Synchronization_Exception);

  /**
   * Destroy this Mutex object
   */
  virtual ~Mutex()
    throw();
  
  /**
   * Acquire a lock on this mutex. 
   *
   * Acquiring this mutex twice will result in deadlock. 
   *
   * @exception Interrupted_Exception 
   * @exception Unexpected_Exception
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   */
  virtual void acquire() 
    throw(Synchronization_Exception);

  /**
   * Acquire a lock on this mutex. 
   *
   * Acquiring this mutex twice will result in deadlock. 
   *
   * @param unsigned long - spend no more than given amount of time waiting 
   * for the mutex
   *
   * @return true if the mutex was acquire in time, false if not
   *
   * @exception Interrupted_Exception 
   * @exception Unexpected_Exception
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   */ 
  virtual bool tryAcquire(unsigned long) 
    throw(Synchronization_Exception);
  
  /**
   * Release a lock held on this mutex. 
   *
   * Once released, another thread may acquire the mutex.
   *
   * @exception Synchronization_Exception
   */
  virtual void release() 
    throw(Synchronization_Exception);
  
}; /* Mutex */


};

#endif
