/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef _ZT_FASTMUTEXIMPL_H_
#define _ZT_FASTMUTEXIMPL_H_

#if !defined(_POSIX)
#error This implementation is specific to POSIX
#endif

#include "zthread/ZThread.h"
#include "zthread/Lockable.h"

#include <pthread.h>

namespace ZThread {

/**
 * @class FastMutex
 *
 * @author Eric Crahen
 * @date 09-21-2000
 * @version 1.1.0
 *
 * This is the smallest and fastest synchronization object in the library.
 * It is an implementation of fast mutex, an all or nothing exclusive
 * lock. It should be used only where you need speed and are willing 
 * to sacrifice all the state & safety checking provided by the framework
 * for speed.
 *
 * There are two important differences from the other Mutex classes. 
 * The first and most important one is speed. It operates about %20 - %25 
 * faster on average than the normal Mutex. 
 *
 */ 
class FastMutex : public Lockable {
    
  pthread_mutex_t _mtx;
  pthread_t _owner;

  // Disable copy constructors
  FastMutex(const FastMutex&);
  FastMutex& operator=(const FastMutex&);
  
  public:
  
  /**
   * Create a new FastMutex. No safety or state checks are performed.
   *
   * @exception Initialization_Exception - not thrown
   */
  inline FastMutex() 
    throw(Synchronization_Exception) : _owner(0) {

    pthread_mutex_init(&_mtx, 0);

  }
  
  /**
   * Destroy a FastMutex. No safety or state checks are performed.
   */
  inline virtual ~FastMutex() 
    throw() {

    pthread_mutex_destroy(&_mtx);

  }
  
  /**
   * Acquire an exclusive lock. No safety or state checks are performed.
   * This is not a recursive function.
   *
   * @exception Synchronization_Exception - not thrown
   */
  virtual void acquire() 
    throw(Synchronization_Exception);

  
  /**
   * Release an exclusive lock. No safety or state checks are performed.
   * The caller should have already acquired the lock, and release it 
   * only once.
   * 
   * @exception Synchronization_Exception - not thrown
   */
  inline virtual void release() 
    throw(Synchronization_Exception);

  
  /**
   * Not implemented
   *
   * @return bool - Always false
   *
   * @exception Synchronization_Exception - not thrown
   */
  inline virtual bool tryAcquire(unsigned long)
    throw(Synchronization_Exception);
    

  
}; /* FastMutex */


};

#endif
