/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef _ZT_POSIXFASTRECUSRIVEMUTEXIMPL_H_
#define _ZT_POSIXFASTRECURSIVEMUTEXIMPL_H_

#if !defined(_POSIX)
#error This implementation is specific to POSIX
#endif

#include "zthread/ZThread.h"
#include "zthread/Lockable.h"

#include <pthread.h>

namespace ZThread {

/**
 * @class FastRecursiveMutex
 *
 * @author Eric Crahen
 * @date 04-22-2000
 * @version 1.3.5
 *
 * This is the smaller and faster implementation of a RecursiveMutex. 
 * A single thread can acquire the mutex any number of times, but it
 * must perform a release for each acquire(). Other threads are blocked
 * until a thread has released all of its locks on the mutex.
 * 
 * This particular implementation performs fewer safety checks. Like 
 * the FastMutex implementation, any waiting caused by an acquire() request 
 * is not interruptable. This is so that the mutex can have the fastest
 * response time for a time critical application while still having a good
 * degree of reliability.
 *
 */ 
class FastRecursiveMutex : public Lockable {
    
  pthread_mutex_t _mtx;
  pthread_t _owner;
  int _count;

  // Disable copy constructors
  FastRecursiveMutex(const FastRecursiveMutex&);
  FastRecursiveMutex& operator=(const FastRecursiveMutex&);

  public:
  
  /**
   * Create a new FastRecursiveMutex
   */
  FastRecursiveMutex() 
    throw(Synchronization_Exception) : _owner(0), _count(0) {

    pthread_mutex_init(&_mtx, 0);

  }

  /**
   * Destroy this Mutex
   */
  inline virtual ~FastRecursiveMutex() throw() {

    pthread_mutex_destroy(&_mtx);

  }

  /**
   * Acquire the lock on this Mutex, if the lock is already held by the
   * calling thread, then the Mutex returns immediately
   *
   * @exception Synchronization_Exception
   */ 
  virtual void acquire() 
    throw(Synchronization_Exception);
   
    
  /**
   * Release the lock on this Mutex. This should be done for each time
   * the Mutex was acquired() by the calling thread
   *
   * @exception Synchronization_Exception
   */ 
  virtual void release() 
    throw(Synchronization_Exception);

    
  /**
   * The timeout function is not implemented on a FastRecursiveMutex
   * This is because this Mutex is generally used when need to be locked
   * and unlocked rather quickly, and where the lock held only for short
   * amounts of time
   *
   * @return bool - always false  
   * @exception Synchronization_Exception
   */ 
  virtual bool tryAcquire(unsigned long)
    throw(Synchronization_Exception);

  
}; /* FastRecursiveMutex */


};

#endif
