/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZPLAINMUTEX_H__
#define __ZPLAINMUTEX_H__

#include "zthread/ZThread.h"
#include "zthread/Lockable.h"

namespace ZThread { 
  
class PlainMutexImpl;

/**
 * @class PlainMutex
 *
 * @author Eric Crahen
 * @version 1.5.0
 * @date 07-31-2001
 *
 * Basic MUTual EXclusion object. Wraps a native mutex providing a minimal amount
 * of functionality.
 *
 * This mutex will perform checks before a wait is started to see if the thread has 
 * been interrupted. However, once a thread is blocked by this mutex it can't be 
 * interrupted. 
 * 
 * The Mutex class provides a mutex implementation that allows interrupts
 * The CheckedMutex provides a mutex with some error checking functionality
 */
class ZTHREAD_API PlainMutex : public Lockable {
  
  PlainMutexImpl* _impl;
  
  // Disable copy constructors
  PlainMutex(const PlainMutex&);
  PlainMutex& operator = (const PlainMutex&);
  
  public:

  /**
   * Allocate a new mutex
   *
   * @exception Initialization_Exception - thrown if there are not
   * enough available resources
   */
  PlainMutex() 
    throw(Synchronization_Exception);

  /**
   * Destroy this PlainMutex object
   */
  virtual ~PlainMutex()
    throw();
  
  /**
   * Acquire a lock on this mutex. If the thread has been interrupted
   * before the lock is a acquired this function will throw an exception
   * If the thread is interrupted after it has started to wait for this
   * mutex it can not be interrupted.
   *
   * Acquiring this mutex twice will result in deadlock. 
   *
   * @exception Interrupted_Exception - the calling thread was interrupted
   * before the lock was obtained - not while waiting for the lock
   * @exception Unexpected_Exception - the calling thread was killed and
   * can't be blocked
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   */
  virtual void acquire() 
    throw(Synchronization_Exception);

  /**
   * Acquire a lock on this mutex. If the thread has been interrupted
   * before the lock is a acquired this function will throw an exception
   * If the thread is interrupted after it has started to wait for this
   * mutex it can not be interrupted.
   *
   * Acquiring this mutex twice will result in deadlock. 
   *
   * @param unsigned long - spend no more than given amount of time waiting 
   * for the mutex
   *
   * @return true if the mutex was acquire in time, false if not
   *
   * @exception Interrupted_Exception - the calling thread was interrupted
   * before the lock was obtained - not while waiting for the lock
   * @exception Unexpected_Exception - the calling thread was killed and
   * can't be blocked
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   */ 
  virtual bool tryAcquire(unsigned long) 
    throw(Synchronization_Exception);
  
  /**
   * Release a lock held on this mutex. Releasing a lock twice is not 
   * guaranteed to throw an InvalidOp exception. This may have an 
   * unpredictable or at least undetected result. It should not be done
   *
   * Once released, another thread may acquire the mutex.
   *
   * No interruption checks are performed in this method
   *
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   */
  virtual void release() 
    throw(Synchronization_Exception);
  
}; /* PlainMutex */


};

#endif
