/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZRECURSIVEMUTEX_H__
#define __ZRECURSIVEMUTEX_H__

#include "zthread/ZThread.h"
#include "zthread/Lockable.h"

namespace ZThread { 
  
class RecursiveMutexImpl;

/**
 * @class RecursiveMutex
 *
 * @author Eric Crahen
 * @version 1.3.4
 * @date 04-12-2001
 *
 * This is an error checking Mutex implementation. Any blocking caused
 * by this mutex can be interrupted. This mutex can also not be deadlocked
 * The performance of this mutex is still fast, but if you really need
 * speed you may want to consider using a FastRecursiveMutex.
 *
 * Each method is an interruption point
 * Each method will throw a meaningful exception
 *
 * You should carefully handle exceptions thrown by this mutex. 
 *
 * One major difference in this class is that the same thread can acquire the
 * lock multiple times. It must release it that same number of times
 *
 * Another is that the release() method will not throw Interrupted_Exceptions
 * The reason for this is to help simplify any nested unlocking that must
 * be performed when a lock is acquire recursively 
 */
class ZTHREAD_API RecursiveMutex : public Lockable {
    
  RecursiveMutexImpl* _impl;
  
  // Disable copy constructors
  RecursiveMutex(const RecursiveMutex&);
  RecursiveMutex& operator=(const RecursiveMutex&);
  
  public:

  /**
   * Create a new CheckedMutex.
   *
   * @exception Initialization_Exception - thrown if resources
   * could not be allocated
   */
  RecursiveMutex() 
    throw(Synchronization_Exception);

  /**
   * Destroy this Mutex
   */
  virtual ~RecursiveMutex()
    throw();

  /**
   * Lock the mutex and prevent re-entrance on the protected block of code.
   * Calling thread is blocked until the Mutex is released by the thread 
   * currently holding its lock
   *
   * @see Loackable::acquire()
   *
   * @exception Interrupted_Exception - thrown if the calling thread was
   * interrupted prior to this call, or during the time it spends waiting
   * to acquire the mutex
   * @exception Deadlock_Exception - NOT thrown because this is a recursive
   * lock.
   * @exception Unexpected_Exception - the calling thread was killed and
   * can't be blocked
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   */ 
  virtual void acquire() 
    throw(Synchronization_Exception);

  /**
   * Lock the mutex and prevent re-entrance on the protected block of code.
   * If the given amount of time expires or the mutex is locked this function
   * returns
   *
   * see Loackable::tryAcquire()
   *
   * @exception Interrupted_Exception - thrown if the calling thread was
   * interrupted prior to this call, or during the time it spends waiting
   * to acquire the mutex
   * @exception Deadlock_Exception - NOT thrown because this is a recursive
   * lock.   
   * @exception Unexpected_Exception - the calling thread was killed and
   * can't be blocked
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   *
   * @return bool indicating the success of the lock attempt
   */ 
  virtual bool tryAcquire(unsigned long) 
    throw(Synchronization_Exception);

  
  /**
   * Unlock the mutex and prevent re-entrance on the protected block of code.
   * This needs to be called for each acquire() that succeeded.
   *
   * see Loackable::release()
   *
   * @post Exceptions are thrown after the release() is performed internally
   *
   * @exception InvalidOp_Exception - thrown if an attempt to release this
   * mutex from a thread that does not own the lock is made
   * @exception Synchronization_Exception - error with the underlying 
   * synchronization mechanisms
   */
  virtual void release() 
    throw(Synchronization_Exception);
  
}; /* RecursiveMutex */
  
};

#endif
