/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZSYNCEXECUTOR_H__
#define __ZSYNCEXECUTOR_H__

#include "zthread/ZThread.h"
#include "zthread/Executor.h"
#include "zthread/Mutex.h"

namespace ZThread {


/**
 * @class SyncExecutor
 *
 * @date 09-21-2000
 * @author Eric Crahen
 * @version 1.3.0
 *
 * This interface defines the basic outline needed to implement 
 * a LWP (light-weigh-process) execution mechanism. 
 *
 * An executor allows the efficient execution of a number of small 
 * unrelated 'Runnable' tasks, or LWP's without the over head of a 
 * large number of synchronization controls needed in a thread per task
 * solution.
 * 
 * The executor itself contains heavy weight synchronization
 * controls to do this, but it will not need new ones for each additional
 * task. This overcomes resource problems thread-per-task designs must face
 *
 * In general, executors queue tasks for execution by another thread or 
 * by a pool of threads. The queuing of task allows for a good deal of 
 * control over the order of execution.
 * 
 * SynchExecutors will never consume tasks. Thier behavior makes it
 * very easy to use shared task objects, or task objects that have been 
 * automatically allocated.
 */
template <class L=Mutex>
class ZTHREAD_API SyncExecutor : public Executor {
  private:

  SyncExecutor(SyncExecutor*);
  SyncExecutor(const SyncExecutor&);
  SyncExecutor& operator=(const SyncExecutor&);
  
  protected:
  
  //! Typedef
  typedef L LOCK;

  //! Serialize access
  LOCK _lock;
  //! Cancellation flag
  bool _canceled;

  public:

  /**
   * Create a new SyncExecutor
   */
  SyncExecutor() : _canceled(false) {}


  /**
   * Destroy this Executor
   */
  virtual ~SyncExecutor() throw() {}

  /**
   * Enqueue another task for execution.
   *
   * @param Runnable task - task to execute
   * @exception InterruptedException 
   *
   * @pre The executor must not be canceled. A canceled
   * executor will ignore execute() requests.
   *
   * @post The given task will be executed as soon as 
   * possible by the executor
   */
  virtual void execute(Runnable* task)
    throw(Synchronization_Exception) {
    
    if(_canceled) // Avoid excess locking in the canceled state
      throw Cancelation_Exception();
    
    Guard<LOCK> g(_lock);
    
    if(_canceled)
      throw Cancelation_Exception();
    
    if(task)
      task->run();

    
  }
  
  /**
   * Stop the execution of additional tasks by this executor.
   * After this method has be invoked then the Executor will
   * completed when the last task it was previously assigned is 
   * completed. 
   *
   * @exception InterruptedException 
   * 
   * @post the executor no longer accepts new tasks. At this point
   * the executor has a limited lifetime and is safe to join
   */
  virtual void cancel() 
    throw(Synchronization_Exception) {
      
    Guard<LOCK> g(_lock);
    _canceled = true;
    
  }
  
  /**
   * Check the cancelation status of this Executor
   * 
   * @return bool - Executor canceled
   * @see Cancelable::isCanceled()
   */
  inline virtual bool isCanceled()
    throw(Synchronization_Exception) {
    
    Guard<LOCK> g(_lock);
    return _canceled;

  }

  /**
   * Wait for this Executor to complete. The calling thread is
   * blocked until the last task has been completed by this executor
   * or until the calling thread has been interrupted.
   *
   * @exception InterruptedException 
   */
  inline virtual void join() 
    throw(Synchronization_Exception) { }

  /**
   * Block until this executor has become available to run new tasks.
   *
   * This does not guarantee that another thread won't have submitted a
   * task before some action is taken in response to this. However, it
   * is useful as a general heuristic
   *
   * @return boolean
   * @exception InterruptedException 
   */
  virtual bool available() 
    throw(Synchronization_Exception) {
      
    Guard<LOCK> g(_lock);
    return !_canceled;

  }

  /**
   * Count of active threads operation on behalf of this executor
   *
   * This does not guarantee that another thread won't have submitted a
   * task that changes the number of executing threads before some action 
   * is taken in response to this. However, it is useful as a general 
   * heuristic
   * 
   * @return unsigned int
   * @exception InterruptedException 
   */ 
  virtual unsigned int threadCount()
    throw(Synchronization_Exception) {

    Guard<LOCK> g(_lock);
    return (!_canceled) ? 1 : 0;

  }


};

};

#endif
