/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZTHREADLOCAL_H__
#define __ZTHREADLOCAL_H__

#include "zthread/ZThread.h"
#include "zthread/AbstractThreadLocal.h"

namespace ZThread {

/**
 * @class ThreadLocal
 *
 * @author Eric Crahen
 * @version 1.1.3
 * @date 01-02-2001
 *
 * Provide a method to access the local storage of each thread. No matter
 * what thread access this object, it will always store values unique to
 * each thread. This is done using the local storage allocated by the OS
 * for each thread. It doesn't require any locking and is very fast
 *
 * The first time a ThreadLocal veriable is accessed by a thread the initialValue()
 * method will be invoked. This allows subclasses to perfrom any special 
 * actions they might need to when a new thread uses one of these variables.
 *
 * The recommended usage of this object would be something like in the
 * example shown below
 *
 * @code
 *
 * class MyClass {
 * protected:
 *
 * static ThreadLocal<int> _threadKey;
 *
 * public:
 * 
 * int getValue() const {
 *   return _threadKey.get();
 * }
 * 
 * int setValue(int n) const {
 *   return _threadKey.set(n);
 * }
 *
 * };
 *
 * @endcode
 *
 * The ThreadLocal object itself acts as the key, instead of some arbitrary
 * integer that needed to be defined by the programmer before. This is a
 * much more elegant solution.
 */
template <class T>
class ZTHREAD_API ThreadLocal : private AbstractThreadLocal {
  public:

  /**
   * Create a new ThreadLocal object
   */
  inline ThreadLocal() throw() {
  }

  /**
   * Destroy this ThreadLocal object
   */
  inline virtual ~ThreadLocal() throw() {
  }

  /**
   * Get a value with this ThreadLocal object. Any value retrieved from
   * this object will have been set from the same Thread. 
   *
   * @exception Synchronization_Exception - thrown if there is an error
   * allocating native thread local storage
   */
  inline T get() const
    throw(Synchronization_Exception) {

    return reinterpret_cast<T>( AbstractThreadLocal::get() );

  }
  
  /**
   * Set a value with this ThreadLocal object. This value can only
   * be retrieved from this ThreadLocal object from the same Thread
   * that set it.
   *
   * @exception Synchronization_Exception - thrown if there is an error
   * allocating native thread local storage
   */
  inline T set(T val) const
    throw(Synchronization_Exception) {

    return reinterpret_cast<T>( AbstractThreadLocal::set((void*)val) );

  }

  protected:

  /**
   * Invoked by the framework the first time a ThreadLocal variable is used 
   * by a thread. Provides a location for subclasses to perform various 
   * per/thread initialization.
   *
   * @param void* initValue - the default value is always 0. Subclasses 
   * may overide this and return a new value or delegate up thier class
   * hierarchy as they see fit. 
   *
   * @return void* - value that will be set for the executing thread
   */
  inline virtual void* initialValue(void* initValue) throw() {   
    return initValue;
  }

  /**
   * Inform the framework this value does not propogate to child threads
   *
   * @return false - always
   */
  inline virtual bool propogateValue() throw() {
    return false;
  }


};


};

#endif
