/*
 *  ZThreads, a platform-independant, multithreading and 
 *  synchroniation library
 *
 *  Copyright (C) 2001 Eric Crahen, See LGPL.TXT for details
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *  SUNY @ Buffalo, hereby disclaims all copyright interest in the
 *  ZThreads library written by Eric Crahen
 */

#ifndef __ZTIMER_H__
#define __ZTIMER_H__

#include "zthread/ZThread.h"
#include "zthread/Cancelable.h"
#include "zthread/SynchronizationExceptions.h"
#include "zthread/Condition.h"
#include "zthread/FastMutex.h"
#include "zthread/NullMutex.h"
#include "zthread/Observable.h"
#include "zthread/Thread.h"

#include <set>

namespace ZThread {

class Timer;
class TimerEvent;

/**
 * @class TimerListener
 *
 * @author Eric Crahen
 * @version 1.1.2
 * @date 12-28-2001
 *
 * An object interested in a Timer can implement this interface and then
 * register itself with a Timer to begin receiving TimerEvents as various
 * lengths of time expire
 */
class ZTHREAD_API TimerListener {
public:

  /**
   * Destroy the TimerListener
   */
  virtual ~TimerListener() throw() {}
 
  /**
   * Called from the Timer thread when an event expires
   *
   * @param const Timer* - timer that waited for the event
   * @param const TimerEvent& - timer event that expired
   */
  virtual void notify(const Timer*, const TimerEvent&) = 0;

};


/**
 * @class TimerEvent
 *
 * @author Eric Crahen
 * @version 1.3.0
 * @date 03-10-2001
 *
 * This class just encapsulates the timing details for each event
 * and makes them a bit easier to work with. 
 */
class ZTHREAD_API TimerEvent {

  unsigned long _interval;
  unsigned long _timeExpired;
  unsigned long _id;
  bool _periodic;

public:

  /**
   * Create a TimerEvent by copying another
   *
   * @param const TimerEvent& - event to copy
   */
  TimerEvent(const TimerEvent&);

  /**
   * Create a new event
   * 
   * @param unsigned long - time at which the timer should expire;
   * milliseconds from now
   * @param unsigned long - timer event id
   * @param bool - flag to indicate if this is a repeating/periodic event
   */
  TimerEvent(unsigned long, unsigned long, bool periodic = false);

  /**
   * Get the event id
   *
   * @return unsigned long - event id
   */
  inline unsigned long getId() const {
    return _id;
  }

  /**
   * Get the time left until this event should occur
   *
   * @return unsigned long - time till event
   */
  unsigned long getExpiration() const;

  /**
   * Get the interval for the event
   *
   * @return unsigned long - interval
   */
  inline unsigned long getInterval() const {
    return _interval;
  }

  /**
   * Check if the event has expired
   *
   * @return bool - event expired if true
   */
  bool isExpired() const;

  /**
   * Check if this is a periodic/repeating event.
   *
   * @return bool - event is periodic/repeating if true
   */
  inline bool isPeriodic() const {
    return _periodic;
  }

  /**
   * Copy an exisitng TimerEvent
   *
   * @param const TimerEvent& - timer event to copy
   */
  TimerEvent& operator=(const TimerEvent&);

  /**
   * Check if the time two events expire is the same
   *
   * @param const TimerEvent& - timer event to compare
   */
  inline bool operator==(const TimerEvent& e) const {
    return _timeExpired == e._timeExpired;
  }

  /**
   * Check if the TimerEvent expired before another
   *
   * @param const TimerEvent& - timer event to compare
   */
  inline bool operator<(const TimerEvent& e) const {
    return _timeExpired < e._timeExpired;
  }

  /**
   * Check if the TimerEvent expired after another
   *
   * @param const TimerEvent& - timer event to compare
   */
  inline bool operator>(const TimerEvent& e) const {
    return _timeExpired > e._timeExpired;
  }

  /**
   * Casting operator to that evaluates to the event id
   */
  inline operator unsigned long() const {
    return getId();
  }

 protected:

  /**
   * Get the current time in milliseconds
   */
  static unsigned long now() throw();

};  


/**
 * @class Timer
 *
 * @author Eric Crahen
 * @version 1.4.1
 * @date 04-28-2001
 *
 * This class defines a simple Timer object base. Other objects can be
 * derived from this to create more specialzed Timers. This encapsulates 
 * the timing routines and doesn't force a timer to have a thread 
 * associated with it. A Timer task could be created and submitted to an
 * execute or manually managed thread.
 *
 * Interest in a timer is manipulated using the inherited Observable methods 
 */
class ZTHREAD_API Timer : public Cancelable, public Runnable, 
  public Observable<TimerListener, NullMutex> {

protected:

  //! Typedef
  typedef std::set<TimerEvent> EventList;
  //! Registered events
  EventList _eventList;
  //! Used to perform the timing
  Condition _interrupt;
  //! Used to serialize access
  FastMutex _lock;

public:

  /**
   * Create a new Timer object
   */
  Timer()
    throw(Synchronization_Exception);

  /**
   * Destroy this Timer object
   */
  virtual ~Timer() throw();

  /**
   * Set a timer event. This event will occur after the given amount of time
   * expires.
   *
   * @param const TimerEvent& - event to submit to the Timer
   *
   * @exception Synchronization_Exception
   */
  void addEvent(const TimerEvent&)
    throw(Synchronization_Exception);

  /**
   * Set a timer. This event will occur after the given amount of time
   * expires. An appropriate TimerEvent will be constructed and submitted
   * to the Timer based on the parameters passed to this function
   *
   * @param unsigned long - milliseconds until event
   * @param bool - event repeats
   *
   * @exception Synchronization_Exception
   */
  void addEvent(unsigned long, bool periodic = false)
    throw(Synchronization_Exception);

  /**
   * Remove a TimerEvent.
   *
   * @param const TimerEvent& - event to remove from the Timer
   *
   * @exception Synchronization_Exception
   */
  void delEvent(const TimerEvent&)
    throw(Synchronization_Exception);

  /**
   * This method enters a loop, waiting on the condition variable that
   * is associated with this object. This loop will executed until
   * the Timer has been canceled.
   */
  virtual void run() throw(); 


}; /* Timer */


};

#endif
