#ifndef __ZWIN32FASTRECURSIVEMUTEXIMPL_H__
#define __ZWIN32FASTRECURSIVEMUTEXIMPL_H__

#if !defined(_WIN32)
#error "Win32 specific implementation"
#endif

#include "zthread/ZThread.h"
#include "zthread/Lockable.h"
#include <windows.h>

namespace ZThread {

/**
 * @class FastRecursiveMutex
 *
 * @author Eric Crahen
 * @date 09-21-2000
 * @version 1.1.0
 *
 * This is the smaller and faster implementation of a RecusiveMutex. 
 * A single thread can aquire the mutex any number of times, but it
 * must perform a release for each acquire(). Other threads are blocked
 * until a thread has released all of its locks on the mutex.
 * 
 * This particular implementation performs fewer safety checks. Like 
 * the FastMutex implementation, any waiting caused by an acquire() request 
 * is not interruptable. This is so that the mutex can have the fastest
 * response time for a time critical application while still having a good
 * degree of reliabililty.
 *
 * The tryAqcuire() method is also not implemented on this mutex, similar
 * to the FastMutex.
 */ 
class ZTHREAD_API FastRecursiveMutex : public Lockable {
    
  CRITICAL_SECTION _csMutex;

  public:
  
  inline FastRecursiveMutex() 
    throw(Synchronization_Exception) {

    ::InitializeCriticalSection(&_csMutex);

  }

  virtual inline ~FastRecursiveMutex() throw()  {
    ::DeleteCriticalSection(&_csMutex);
  }

  virtual inline void acquire() 
    throw(Synchronization_Exception) {

    ::EnterCriticalSection(&_csMutex);      

  }
    
    
  virtual inline void release() 
    throw(Synchronization_Exception) {

    ::LeaveCriticalSection(&_csMutex);

  }
    
  virtual bool tryAcquire(unsigned long)
    throw(Synchronization_Exception);
  
}; /* FastRecursiveMutex */


};



#endif
  
