//
// This example demonstrates the use of sempahores to build a simple
// resource queue. The procuder should fill the queue faster than the
// consumer can empty it - stopping the producer when it catches a 
// range exceeded exception. Then the consumer will continue to 
// empty the queue - after 20 seconds the example quits
//
#include <zthread/Thread.h>
#include <zthread/Semaphore.h>
#include <zthread/FastMutex.h>
#include <zthread/Guard.h>

#include <iostream>
#include <list>

using namespace ZThread;
using namespace std;

// Bounded resource queue, starts empty - max 5 items

class Resources {

  FastMutex lock;
  Semaphore* s;
  list<int> data;

public:

  Resources() 
    throw(Initialization_Exception) {

    s = new Semaphore(0, 5);

  }

  ~Resources() {

    delete s;

  }

  int get()
    throw(Synchronization_Exception) {

    int val;
    s->acquire();

    Guard<FastMutex> guard(lock);

    val = data.front(); data.pop_front();

    return val;

  }

  void put(int val)
    throw(Synchronization_Exception) {

    s->release();
    
    Guard<FastMutex> guard(lock);
    data.push_back(val);

  }

  int count()
    throw(Synchronization_Exception) {

    Guard<FastMutex> guard(lock);
    return s->count();

  }

};


static Resources pool;

class Consumer : public Runnable {
public:

  virtual ~Consumer() {}

  virtual void run() throw() {

    try {

      for(;;) {
      
        cout << "GET: " << pool.get() << endl;
        Thread::sleep(2000);

      }

    } catch(Synchronization_Exception& e) { 
      cout << "Consumer exception: " << e.what() << endl; 
    }

  }

};

class Producer : public Runnable {
public:

  virtual ~Producer() {}

  virtual void run() throw() {

    try {

      for(;;) {
      
        int val = (rand() % 10) + 1;

        cout << "PUT: " << val << endl;
        pool.put(val);

        Thread::sleep(1500);

      }

    } catch(Synchronization_Exception& e) { 
      cout << "Producer exception: " << e.what() << endl; 
    }

  }

};

int main(int argc, char* argv[]) {

  try {

    Consumer cTask;
    Thread c;
    c.run( &cTask );

    Producer pTask;
    Thread p;
    p.run( &pTask );

    cout << "Main thread, sleeping 20 seconds" << endl;

    Thread::sleep(20000);

    c.interrupt();
    p.interrupt();

    c.join();
    p.join();

  } catch(Synchronization_Exception& e) { cout << e.what() << endl; }

#ifdef WIN32
  cout << "Press a key" << endl;
  getchar(); // Otherwise the console window will close to quickly
#endif
  return 0;

}
